"""Hillstrom challenge dataset.

A dataset from Kevin Hillstrom's MineThatData blog. See

    https://blog.minethatdata.com/2008/03/minethatdata-e-mail-analytics-and-data.html

for details.
"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="Hillstrom.csv",
    url=('http://www.minethatdata.com/'
         'Kevin_Hillstrom_MineThatData_E-MailAnalytics'
         '_DataMiningChallenge_2008.03.20.csv'),
    checksum=('0e5893329d8b93cefecc571777672028'
              '290ab69865718020c78c7284f291aece'))


def fetch_Hillstrom(data_home=None, download_if_missing=True,
                    random_state=None, shuffle=False,
                    categ_as_strings=False, return_X_y=False,
                    as_frame=False):
    """Load the Hillstrom dataset (uplift classification and regression).

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_visit : numpy array
        Each value is 1 if website visit occurred 0 otherwise.

    dataset.target_conversion : numpy array
        Each value is 1 if purchase occurred 0 otherwise.

    dataset.target_spend : numpy array
        Each value corresponds to the amount of money spent.

    dataset.DESCR : string
        Description of the Hillstrom dataset.

    (data, target_visit, target_conversion, target_spend) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['No E-Mail', 'Mens E-Mail', 'Womens E-Mail']
    history_segment_values = ['1) $0 - $100', '2) $100 - $200',
                              '3) $200 - $350', '4) $350 - $500',
                              '5) $500 - $750', '6) $750 - $1,000',
                              '7) $1,000 +']
    zip_code_values = ['Rural', 'Surburban', 'Urban']
    channel_values = ['Phone', 'Web', 'Multichannel']
    categ_values = {"history_segment": history_segment_values,
                    "zip_code": zip_code_values,
                    "channel": channel_values,}

    # attribute descriptions
    treatment_descr = [("segment", treatment_values)]
    target_descr = [("target_visit", np.int32, "visit"),
                    ("target_conversion", np.int32, "conversion"),
                    ("target_spend", float, "spend")]
    feature_descr = [("recency", np.int32),
                     ("history_segment", history_segment_values),
                     ("history", float),
                     ("mens", np.int32),
                     ("womens", np.int32),
                     ("zip_code", zip_code_values),
                     ("newbie", np.int32),
                     ("channel", channel_values)]

    ret = _fetch_remote_csv(ARCHIVE, "Hillstrom",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=12
                            )
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""Criteo online advertising dataset.

See https://ailab.criteo.com/criteo-uplift-prediction-dataset/ for details.
"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


# original source
#ARCHIVE = RemoteFileMetadata(
#    filename="criteo-research-uplift-v2.1.csv.gz",
#    url=('https://go.criteo.net/criteo-research-uplift-v2.1.csv.gz'),
#    checksum=('2716e1bf0fd157a93b5bf86924d90884'
#              '19dfbac2022c6cd90030220634f616dc'))
# GitHub copy, seems faster
ARCHIVE = RemoteFileMetadata(
    filename="criteo-research-uplift-v2.1.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/releases/download/Criteo/criteo-research-uplift-v2.1.csv.gz'),
    checksum=('2716e1bf0fd157a93b5bf86924d90884'
              '19dfbac2022c6cd90030220634f616dc'))


def fetch_Criteo(data_home=None, download_if_missing=True,
                    random_state=None, shuffle=False,
                    categ_as_strings=False, return_X_y=False,
                    as_frame=False):
    """Load the Criteo dataset.

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_visit : numpy array
        Each value is 1 if website visit occurred 0 otherwise.

    dataset.target_conversion : numpy array
        Each value is 1 if purchase occurred 0 otherwise.

    dataset.target_spend : numpy array
        Each value corresponds to the amount of money spent.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_conversion, target_visit, target_exposure) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['control', 'treated']
    categ_values = dict()

    # attribute descriptions
    treatment_descr = [("treatment", np.int32)]
    target_descr = [("target_conversion", np.int32, "conversion"),
                    ("target_visit", np.int32, "visit"),
                    ("target_exposure", np.int32, "exposure")]
    feature_descr = [(f"f{i}", float) for i in range(12)]

    ret = _fetch_remote_csv(ARCHIVE, "Criteo",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=16, all_num=True
                            )
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""
The :mod:`usklearn.model_selection._search` includes utilities to fine-tune the
parameters of an estimator.
"""

from collections.abc import Mapping, Sequence, Iterable
from functools import partial, reduce
from itertools import product
from copy import deepcopy
import numbers

import numpy as np

from sklearn.utils import get_tags
from sklearn.base import BaseEstimator, is_classifier, clone
from sklearn.utils.validation import indexable, check_is_fitted
from sklearn.metrics import check_scoring

from ..metrics import check_uplift_scoring

from ._validation import uplift_check_cv
from ._validation import _check_multimetric_scoring


__all__ = ['GridSearchCV', 'ParameterGrid', 'fit_grid_point',
           'ParameterSampler', 'RandomizedSearchCV']

from collections.abc import Mapping, Sequence, Iterable

from sklearn.model_selection import GridSearchCV as _sklearn_GridSearchCV

from ..utils import check_trt
from ..utils import MultiArray

from ._validation import _WrappedUpliftEstimator, _WrappedScoring


class BaseSearchCV(BaseEstimator):
    @property
    def _estimator_type(self):
        return self.estimator._estimator_type
    def __sklearn_tags__(self):
        # Copied from sklearn 1.6.1
        tags = super().__sklearn_tags__()
        sub_estimator_tags = get_tags(self.estimator)
        tags.estimator_type = sub_estimator_tags.estimator_type
        tags.classifier_tags = deepcopy(sub_estimator_tags.classifier_tags)
        tags.regressor_tags = deepcopy(sub_estimator_tags.regressor_tags)
        # allows cross-validation to see 'precomputed' metrics
        tags.input_tags.pairwise = sub_estimator_tags.input_tags.pairwise
        tags.input_tags.sparse = sub_estimator_tags.input_tags.sparse
        tags.array_api_support = sub_estimator_tags.array_api_support
        return tags

    def __getattr__(self, arg):
        """Delegate to real grid search."""
        if "wrapped_search_" in self.__dict__:
            return getattr(self.wrapped_search_, arg)
        raise AttributeError(arg)
    def fit(self, X, y, trt, n_trt=None, **fit_params):
        X, y, trt = indexable(X, y, trt)
        trt, n_trt = check_trt(trt, n_trt)

        wrapped_est = _WrappedUpliftEstimator(self.estimator)
        scoring = check_uplift_scoring(self.estimator, self.scoring)
        wrapped_scoring = _WrappedScoring(scoring)
        # TODO: multimetric scoring
        #scorers, _ = _check_multimetric_scoring(estimator, scoring=scoring)
        #wrapped_scorers = {name:_WrappedScoring(scoring) for name, scoring in scorers.items()}
        # TODO: fix cv
        cv, y_stratify = uplift_check_cv(self.cv, y, trt, n_trt, classifier=is_classifier(self.estimator))
        # fix param_grid for wrapped uplift estimator
        param_grid = self.param_grid
        if isinstance(self.param_grid, Mapping):
            param_grid = [param_grid]
        new_param_grid = []
        for grid in param_grid:
            new_grid = {"base_estimator__" + key:value for key, value in grid.items()}
            new_param_grid.append(new_grid)
        self.wrapped_search_ = self.wrapped_search_class(estimator=wrapped_est, param_grid=new_param_grid, scoring=wrapped_scoring,
                                                    cv=cv, n_jobs=self.n_jobs, refit=self.refit, verbose=self.verbose,
                                                    pre_dispatch=self.pre_dispatch, error_score=self.error_score,
                                                        return_train_score=self.return_train_score)

        # multiarray to pass additional data
        # y_stratify is used only for stratification
        Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
        self.wrapped_search_.fit(Xm, y_stratify, **fit_params)
    def score(self, X, y, trt, n_trt=None):
        Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
        return self.wrapped_search_.score(Xm, y)

class GridSearchCV(BaseSearchCV):
    wrapped_search_class = _sklearn_GridSearchCV
    def __init__(self, estimator, param_grid, *, scoring=None,
                 n_jobs=None, refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', error_score=np.nan,
                 return_train_score=False):
        self.estimator = estimator
        self.param_grid = param_grid
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.refit = refit
        self.cv = cv
        self.verbose = verbose
        self.pre_dispatch = pre_dispatch
        self.error_score = error_score
        self.return_train_score = return_train_score




# TODO:
class RandomizedSearchCV(BaseEstimator):
    _required_parameters = ["estimator", "param_distributions"]

    def __init__(self, estimator, param_distributions, *, n_iter=10, scoring=None,
                 n_jobs=None, refit=True,
                 cv=None, verbose=0, pre_dispatch='2*n_jobs',
                 random_state=None, error_score=np.nan,
                 return_train_score=False):
        self.param_distributions = param_distributions
        self.n_iter = n_iter
        self.random_state = random_state
        super().__init__(
            estimator=estimator, scoring=scoring,
            n_jobs=n_jobs, iid=iid, refit=refit, cv=cv, verbose=verbose,
            pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)

    def _run_search(self, evaluate_candidates):
        """Search n_iter candidates from param_distributions"""
        evaluate_candidates(ParameterSampler(
            self.param_distributions, self.n_iter,
            random_state=self.random_state))



import sklearn.pipeline
from sklearn.utils.metaestimators import if_delegate_has_method

class Pipeline(sklearn.pipeline.Pipeline):
    """Uplift version of sklearn's Pipeline.

    allows trt, n_trt args in fit method and passes them to the final
    estimator. (it makes no sense to pass them to transforms since trt
    is not available for future data).
    """
    
    def fit(self, X, y=None, trt=None, n_trt=None, **fit_params):
        n, _ = self.steps[-1]
        fit_params[n + "__trt"] = trt
        fit_params[n + "__n_trt"] = n_trt
        super().fit(X, y, **fit_params)
    # a copy of score method was necessary since it does not accept
    # score_params as arguments
    @if_delegate_has_method(delegate='_final_estimator')
    def score(self, X, y=None, trt=None, n_trt=None, sample_weight=None):
        """Apply transforms, and score with the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        y : iterable, default=None
            Targets used for scoring. Must fulfill label requirements for all
            steps of the pipeline.
        trt : vector of treatments
        n_trt : number of different treatments
        sample_weight : array-like, default=None
            If not None, this argument is passed as ``sample_weight`` keyword
            argument to the ``score`` method of the final estimator.

        Returns
        -------
        score : float
        """
        
        Xt = X
        for name, transform in self.steps[:-1]:
            if transform is not None:
                Xt = transform.transform(Xt)
        score_params = {}
        if sample_weight is not None:
            score_params['sample_weight'] = sample_weight
        return self.steps[-1][-1].score(Xt, y, trt, n_trt, **score_params)
    @if_delegate_has_method(delegate='_final_estimator')
    def predict_action(self, X, y=None, **predict_action_params):
        """Apply transforms, and score with the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.
        y : iterable, default=None
            Targets used for scoring. Must fulfill label requirements for all
            steps of the pipeline.
        predict_action_params : additional parameters to pass to
            predict_action method of final estimator

        Returns
        -------
        score : float
        """
        Xt = X
        for name, transform in self.steps[:-1]:
            if transform is not None:
                Xt = transform.transform(Xt)
        return self.steps[-1][-1].predict_action(Xt,
                                        **predict_action_params)



"""The X-learner model from Künzel et al. 'Metalearners for estimating
heterogeneous treatment effects using machine learning'.

Currently available only for regression.  Can be applied to
classification problems when treating class variable as numeric.

"""

import numpy as np

from sklearn.linear_model import LinearRegression

from .base import UpliftMetaModelBase
from ..base import UpliftRegressorMixin

class XLearnerUpliftRegressor(UpliftRegressorMixin, UpliftMetaModelBase):
    def __init__(self, base_estimator = LinearRegression()):
        super().__init__(base_estimator=base_estimator)
    def _get_model_names_list(self, X=None, y=None, trt=None):
        if self.n_trt_ > 1:
            raise ValueError("XLearner is only supported for single treatments")
        m_names = ["model_c", "model_t", "model_c_hat", "model_t_hat"]
        return m_names
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        for i in range(self.n_trt_ + 1):
            mask = (trt==i)
            if sample_weight is None:
                wi = None
            else:
                wi = sample_weight[mask]
            Xi = X[mask]
            yi = y[mask]
            yield Xi, yi, wi
        c_mask = (trt==0)
        t_mask = (trt==1)
        # c_hat model
        if sample_weight is None:
            wi = None
        else:
            wi = sample_weight[c_mask]
        y_c_pred = self.models_[1][1].predict(X[c_mask])
        yield X[c_mask], y_c_pred - y[c_mask], wi
        # t_hat model
        if sample_weight is None:
            wi = None
        else:
            wi = sample_weight[t_mask]
        y_t_pred = self.models_[0][1].predict(X[t_mask])
        yield X[t_mask], y[t_mask] - y_t_pred, wi
    def predict(self, X):
        n = self.n_[0] + self.n_[1]
        pc = self.n_[0] / n
        pt = self.n_[1] / n
        pred_c_hat = self.models_[2][1].predict(X)
        pred_t_hat = self.models_[3][1].predict(X)
        y = pc * pred_c_hat + pt * pred_t_hat
        return y



"""The Lalonde datasets.

There are two versions of the dataset, including 1974 earnings
(version B) and excluding 1974 earnings (vestion A).
"""


import logging
from os.path import dirname, exists, join
from os import remove, makedirs
import csv

import numpy as np

from sklearn.datasets import get_data_home
from .base import _fetch_remote_csv
from .base import _prepare_final_data
from .base import RemoteFileMetadata
from sklearn.utils import Bunch
import joblib
from sklearn.utils import check_random_state

ARCHIVE_A_T = RemoteFileMetadata(
    filename="Lalonde_A_T.txt",
    url='http://www.nber.org/~rdehejia/data/nsw_treated.txt',
    checksum=('ab65cd58de17a78b692e66e4d7142192'
              '59ac180428f24c42ddbb928cfb1820fe'))
ARCHIVE_A_C = RemoteFileMetadata(
    filename="Lalonde_A_C.txt",
    url='http://www.nber.org/~rdehejia/data/nsw_control.txt',
    checksum=('8fd745ed2c3426bb77e34b395fb84d45'
              '6d346ba545af58a65b4963160b0699fd'))

ARCHIVE_B_T = RemoteFileMetadata(
    filename="Lalonde_B_T.txt",
    url='http://www.nber.org/~rdehejia/data/nswre74_treated.txt',
    checksum=('e7b742fe0ff07a0f45e129b4ff108bb9'
              '611cd83d53604732c48a8a0a3e20eda3'))
ARCHIVE_B_C = RemoteFileMetadata(
    filename="Lalonde_B_C.txt",
    url='http://www.nber.org/~rdehejia/data/nswre74_control.txt',
    checksum=('a1364cea459d953dc691a667d99194b4'
              'ad335d6d550354fe23a5d2dc58d729b5'))

logger = logging.getLogger(__name__)


def fetch_Lalonde(version="A", data_home=None,
                  categ_as_strings=False,
                  download_if_missing=True, random_state=None,
                  shuffle=False, return_X_y=False,
                  as_frame=False):
    """Load the Lalonde datasets (uplift regression).

    Download it if necessary.

    There are two versions of the dataset, including 1974 earnings
    (version B) and excluding 1974 earnings (vestion A).
    Source: http://users.nber.org/~rdehejia/data/nswdata2.html

    Parameters
    ----------
    version : string, optional
        Specify which dataset to return.  ``A'' for larger files
        without 1974 earnings, ``B'' for smaller files with 1974
        earnings.

    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.

    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array of shape (581012, 54)
        Each row corresponds to the 54 features in the dataset.

    dataset.target : numpy array of shape (581012,)
        Each value corresponds to one of the 7 forest covertypes with values
        ranging between 1 to 7.

    dataset.DESCR : string
        Description of the forest covertype dataset.

    (data, target) : tuple if ``return_X_y`` is True

    """

      
    #data_home = get_data_home(data_home=data_home)
    #Lalonde_dir = join(data_home, "uplift_sklearn", "Lalonde")
    #samples_path = join(Lalonde_dir, "samples" + version_suffix)
    #targets_path = join(Lalonde_dir, "targets" + version_suffix)
    #treatment_path = join(Lalonde_dir, "treatment" + version_suffix)
    #available = exists(samples_path)

    # dictionaries
    header_A = ['treatment', 'age', 'education', 'Black', 'Hispanic',
                'married', 'nodegree', 'RE75', 'RE78']
    header_B = ['treatment', 'age', 'education', 'Black', 'Hispanic',
                'married', 'nodegree', 'RE74', 'RE75', 'RE78']

    target_names = ["RE78"]

    def _float_to_int(x):
        return np.array(x, float), np.int32
    
    # attribute descriptions
    treatment_descr = [("treatment", _float_to_int)]
    target_descr = [("target_RE78", float, "RE78")]
    feature_descr_all = [("age", float),
                         ("education", float),
                         ("Black", float),
                         ("Hispanic", float),
                         ("married", float),
                         ("nodegree", float)]
    feature_descr_A = feature_descr_all + [("RE75", float)]
    feature_descr_B = feature_descr_all + [("RE74", float), ("RE75", float)]
    csv_reader_args = {"delimiter":' ', "skipinitialspace":True}
    
    # choose version
    if version == "A":
        version_suffix = "_A"
        arch_t = ARCHIVE_A_T
        arch_c = ARCHIVE_A_C
        n_fields = 9
        header = header_A
        feature_descr = feature_descr_A
    elif version == "B":
        version_suffix = "_B"
        arch_t = ARCHIVE_B_T
        arch_c = ARCHIVE_B_C
        n_fields = 10
        header = header_B
        feature_descr = feature_descr_B
    else:
        raise ValueError("Lalonde dataset version must be A or B")



    D_T = _fetch_remote_csv(arch_t, "Lalonde_T"+version_suffix,
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=False, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=False,
                            total_attrs=n_fields, header=header,
                            csv_reader_args=csv_reader_args
                            )
    assert np.all(D_T.treatment == 1)
    D_C = _fetch_remote_csv(arch_c, "Lalonde_C"+version_suffix,
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=False, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=False,
                            total_attrs=n_fields, header=header,
                            csv_reader_args=csv_reader_args
                            )
    assert np.all(D_C.treatment == 0)

    # combine treatment and control datasets
    D = D_C
    if as_frame:
        import pandas
        D.data = pandas.concat([D_C.data, D_T.data], ignore_index=True)
    else:
        D.data = np.concatenate([D_C.data, D_T.data])
    D.treatment = np.concatenate([D_C.treatment, D_T.treatment])
    D.target_RE78 = np.concatenate([D_C.target_RE78, D_T.target_RE78])

    # prepare final return value
    ret = _prepare_final_data(D, shuffle, return_X_y)
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""Base classes for all estimators and trasformers."""

import numpy as np

from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.utils.multiclass import unique_labels

from .metrics import e_sate

class _BaseUpliftMixin:
    def predict_action(self, X):
        """Predict most beneficial action."""
        y = self.predict(X)
        if self.n_trt_ == 1:
            a = (y > 0)*1
        else:
            a = np.argmax(y, axis=1) + 1
            best_y = np.max(y, axis=1)
            a[best_y <= 0] == 0
        return a
    def _set_fit_params(self, y, trt, n_trt):
        """Set attributes related to fitted data common to all models
        of a given type."""
        self.trt_, self.n_trt_ = trt, n_trt

class UpliftRegressorMixin(_BaseUpliftMixin, RegressorMixin):
    _estimator_type = "regressor"
    _uplift_model = True

    def score(self, X, y, trt, n_trt=None, sample_weight=None):
        return -e_sate(y, self.predict(X), trt, n_trt=self.n_trt_)

class UpliftClassifierMixin(_BaseUpliftMixin, ClassifierMixin):
    _estimator_type = "classifier"
    _uplift_model = True

    def predict_action(self, X, pos_label=None):
        """Predict most beneficial action.

        Only supported for binary classification or when pos_label is
        set.  pos_label must be an intereger between 0 and
        self.n_classes_-1.

        """
        if pos_label is None and self.n_classes_ > 2:
            raise RuntimeError("predict_action only available for"\
                               "binary classifiers or when positive label is explicitly"\
                               "specified.")
        if pos_label is None:
            pos_label = 1
        try:
            range(self.n_classes_)[pos_label]
            assert pos_label >= 0
        except (IndexError, AssertionError):
            raise ValueError("pos_label must be an intereger between 0 and"\
                             "self.n_classes_-1.")
        y = self.predict(X)
        if self.n_trt_ == 1:
            a = (y[:,pos_label] > 0)*1
        else:
            a = np.argmax(y, axis=1) + 1
            best_y = np.max(y, axis=1)
            a[best_y <= 0] == 0
        return a
    def _set_fit_params(self, y, trt, n_trt):
        """Also set class list and number of classes."""
        super()._set_fit_params(y, trt, n_trt)
        self.classes_ = unique_labels(y)
        self.n_classes_ = len(self.classes_)
    def score(self, X, y, trt, n_trt=None, sample_weight=None):
        return -e_sate(y, self.predict(X), trt, n_trt=self.n_trt_)


class UpliftTransformerMixin(object):
    def fit_transform(self, X, y=None, trt=None, n_trt=None, **fit_params):
        # non-optimized default implementation; override when a better
        # method is possible for a given clustering algorithm
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            fitted = self.fit(X, trt, n_trt, **fit_params)
            return fitted.transform(X, trt, n_trt)
        else:
            # fit method of arity 2 (supervised transformation)
            fitted = self.fit(X, y, trt, n_trt, **fit_params)
            return fitted.transform(X, trt, n_trt)

def is_uplift(estimator):
    """Returns True if the given estimator is an uplift model.

    Parameters
    ----------
    estimator : object
        Estimator object to test.

    Returns
    -------
    out : bool
        True if estimator is an uplift model and False otherwise.
    """
    return getattr(estimator, "_uplift_model", None) is True



"""'Fake' uplift models based on response classifiers."""

import numpy as np

from sklearn.linear_model import LogisticRegression

from .multi_model import _MultimodelUpliftClassifierBase

class _ResponseModelBase(_MultimodelUpliftClassifierBase):
    def __init__(self, base_estimator, reverse, ignore_control):
        super().__init__(base_estimator=base_estimator,
                         ignore_control=ignore_control)
        self.reverse = reverse
    def predict(self, X):
        y = super().predict(X)
        if self.reverse:
            y = 1-y
        return y

class TreatmentUpliftClassifier(_ResponseModelBase):
    """Predict uplift based on treatment classifiers.

    Ignore control."""
    def __init__(self, base_estimator=LogisticRegression(), reverse=False):
        super().__init__(base_estimator, reverse=reverse,
                         ignore_control=True)
class ResponseUpliftClassifier(TreatmentUpliftClassifier):
    """Predict uplift using a classifier built on full data.

    Ignore causal nature of the data."""
    def __init__(self, base_estimator=LogisticRegression(), reverse=False):
        super().__init__(base_estimator, reverse=reverse)
    def fit(self, X, y, trt, n_trt=None, sample_weight=None):
        n = X.shape[0]
        trt = np.ones(n, dtype=np.int32)
        n_trt = 1
        super().fit(X, y, trt, n_trt, sample_weight=sample_weight)
class ControlUpliftClassifier(TreatmentUpliftClassifier):
    """Predict uplift based on a control classifier.

    Ignore treatment data.  If reverse is True lower classification
    scores are assumed to correspond to higher uplift.

    """
    def __init__(self, base_estimator=LogisticRegression(), reverse=True):
        super().__init__(base_estimator=base_estimator, reverse=reverse)
    def fit(self, X, y, trt, n_trt=None, sample_weight=None):
        mask = (trt == 0)
        n = mask.sum()
        trt = np.ones(n, dtype=np.int32)
        n_trt = 1
        if sample_weight is not None:
            sample_weight = sample_weight[mask]
        super().fit(X[mask], y[mask], trt, n_trt, sample_weight=sample_weight)




"""
The :mod:`usklearn.datasets` module includes utilities to load
popular uplift reference datasets. It also features some artificial
data generators.
"""

from .Hillstrom import fetch_Hillstrom
from .Lalonde import fetch_Lalonde
from .TelecomChurn import fetch_TelecomChurn
from .Criteo import fetch_Criteo
from .Lenta import fetch_Lenta
from .colon import fetch_colon
from .Starbucks import fetch_Starbucks
from .Megafon import fetch_Megafon
from .Lazada import fetch_Lazada
from .Information import fetch_Information
from .MarketingAB import fetch_marketing_AB



"""Utilities for all datasets.

Essentially copied from sklearn.datasets._base to reduce dependency on
unofficial sklearn API."""

from collections import namedtuple, OrderedDict
import hashlib
from urllib.request import urlretrieve
from os.path import join, exists
from os import remove, makedirs
import csv
import gzip
import logging
from inspect import isfunction

import numpy as np

import joblib
from sklearn.utils import Bunch
from sklearn.utils import check_random_state
from sklearn.datasets import get_data_home

RemoteFileMetadata = namedtuple('RemoteFileMetadata',
                                ['filename', 'url', 'checksum'])


def _sha256(path):
    """Calculate the sha256 hash of the file at path."""
    sha256hash = hashlib.sha256()
    chunk_size = 8192
    with open(path, "rb") as f:
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)
    return sha256hash.hexdigest()

# TODO: replace with sklearn.fetch_remote when it becomes available in stable
def _fetch_remote(remote, dirname=None):
    """Helper function to download a remote dataset into path

    Fetch a dataset pointed by remote's url, save into path using remote's
    filename and ensure its integrity based on the SHA256 Checksum of the
    downloaded file.

    Parameters
    ----------
    remote : RemoteFileMetadata
        Named tuple containing remote dataset meta information: url, filename
        and checksum

    dirname : string
        Directory to save the file to.

    Returns
    -------
    file_path: string
        Full path of the created file.
    """

    file_path = (remote.filename if dirname is None
                 else join(dirname, remote.filename))
    urlretrieve(remote.url, file_path)
    checksum = _sha256(file_path)
    if remote.checksum != checksum:
        raise IOError("{} has an SHA256 checksum ({}) "
                      "differing from expected ({}), "
                      "file may be corrupted.".format(file_path, checksum,
                                                      remote.checksum))
    return file_path

def _read_csv(archive_path, feature_attrs, treatment_attrs, target_attrs,
              total_attrs=None, categ_as_strings=False, header=None,
              csv_reader_args={"delimiter":",", "quotechar":'"'},
              all_num=False):
    """Read CSV data.

    feature_attrs, treatment_attrs, target_attrs contain descriptions
    of resp. predictive features, treatment description, and targets.
    Currently only a single treatment attribute is supported.  Each
    description is a list whose elements are tuples describing each
    attribute.  The first element of the tuple is attribute's name,
    the second its type, third element (optional) is the name of
    attribute in the CVS header.  If the type is a list, it is assumed
    to be a list of categories.  If the type is a dict, it is assumed
    that its keys are categories present in the csv file and the dict
    is used to change category names.  If the type is a function, it
    will receive as argument a list of columns values and should
    return a transformed list and final numpy dtype.  Otherwise, type
    should be a valid numpy dtype.
    
    If total_attrs is not None, it should contain the total number of
    attributes in each record.

    if all_num is True, array is read as floats for speed.  This
    should be set only if all columns are either floats or ints.

    """
    def parse_attr(Xy_columns, header, attr_description, categ_as_strings):
        """Parse a single attribute.

        Return values and dtype.  Handle categorical attributes correctly."""
        if len(attr_description) == 2:
            attr_name, attr_dtype = attr_description
            file_attr_name = attr_name
        elif len(attr_description) == 3:
            attr_name, attr_dtype, file_attr_name = attr_description
        attr_no = header.index(file_attr_name)
        x = Xy_columns[attr_no]
        if isinstance(attr_dtype, (list, dict)):
            if categ_as_strings:
                if isinstance(attr_dtype, list):
                    categs = {c:c for c in attr_dtype}
                else: # dict
                    categs = attr_dtype
                maxlen = max(len(c) for c in attr_dtype.values())
                attr_dtype = f"U{maxlen}"
            else:
                categs = {c:i for i, c in enumerate(attr_dtype)}
                attr_dtype = np.int32
            x = [categs[c] for c in x]
        elif isfunction(attr_dtype):
            x, attr_dtype = attr_dtype(x)
        x = np.array(x, dtype=attr_dtype)
        return x, attr_name

    Xy = []
    if archive_path.endswith(".gz"):
        f_open = gzip.open
    else:
        f_open = open
    with f_open(archive_path, mode="rt") as csvfile:
        if header is None:
            delim = csv_reader_args.get("delimiter", ",")
            header = next(csvfile).strip().split(delim)
        if all_num:
            Xy_columns = np.loadtxt(csvfile, unpack=True, **csv_reader_args)
            if total_attrs is not None:
                assert len(Xy_columns) == total_attrs, (record, total_attrs)
        else:
            csvreader = csv.reader(csvfile, **csv_reader_args)
            for record in csvreader:
                Xy.append(record)
                if total_attrs is not None:
                    assert len(record) == total_attrs, (record, total_attrs)
            Xy_columns = list(zip(*Xy))
    remove(archive_path)
    
    # parse treatment
    if len(treatment_attrs) != 1:
        raise RuntimeError("Only one treatment is supported")
    trt, attr_name = parse_attr(Xy_columns, header, treatment_attrs[0], categ_as_strings)
    if isinstance(treatment_attrs[0][1], list):
        treatment_values = list(treatment_attrs[0][1])
    else:
        treatment_values = [str(i) for i in range(max(trt) + 1)]

    # parse targets
    targets = OrderedDict()
    for a_descr in target_attrs:
        ta, attr_name = parse_attr(Xy_columns, header, a_descr, categ_as_strings)
        targets[attr_name] = ta
    target_names = list(targets.keys())

    # predictors
    feature_names = []
    columns = []
    for a_descr in feature_attrs:
        fa, attr_name = parse_attr(Xy_columns, header, a_descr, categ_as_strings)
        feature_names.append(attr_name)
        columns.append(fa)
    categ_values = dict()
    for a in feature_attrs:
        if isinstance(a[1], list):
            categ_values[a[0]] = list(a[1])
        elif isinstance(a[1], dict):
            categ_values[a[0]] = list(a[1].values())
    #X = np.core.records.fromarrays(columns, names=feature_names)

    # create a Bunch
    ret = Bunch(data=columns, treatment=trt,
                feature_names=feature_names,
                target_names=target_names)
    for attr_name in targets:
        ret[attr_name] = targets[attr_name]
    ret.treatment_values=treatment_values
    ret.n_trt=len(treatment_values)-1
    ret.categ_values=categ_values
    return ret

def _prepare_final_data(D, shuffle, return_X_y):
    if shuffle:
        ind = np.arange(D.data.shape[0])
        rng = check_random_state(random_state)
        rng.shuffle(ind)
        D.data = D.data[ind]
        D.treatment = D.treatment[ind]
        for ta in D.target_names:
            D[ta] = D[ta][ind] 

    if return_X_y:
        X = D.data
        targets = tuple(D[tn] for tn in D.target_names)
        trt = D.trt
        ret = (X,) + targets + (trt,)
    else:
        ret = D
    return ret

def _fetch_remote_csv(remote, dataset_name,
                      feature_attrs, treatment_attrs, target_attrs,
                      categ_as_strings=False, return_X_y=False,
                      as_frame=False,
                      download_if_missing=True,
                      random_state=None, shuffle=False,
                      header=None, total_attrs=None,
                      csv_reader_args={"delimiter":",", "quotechar":'"'},
                      all_num=False,
                      data_home=None, logger=None):
    if logger is None:
        logger = logging.getLogger(__name__ + "." + dataset_name)
    data_home = get_data_home(data_home=data_home)
    dataset_dir = join(data_home, "uplift_sklearn", dataset_name)
    if categ_as_strings:
        dataset_path = join(dataset_dir, "bunch_str")
    else:
        dataset_path = join(dataset_dir, "bunch")
    available = exists(dataset_path)

    if not available and download_if_missing:
        if not exists(dataset_dir):
            makedirs(dataset_dir)
        logger.info("Downloading %s" % remote.url)
        archive_path = _fetch_remote(remote, dirname=dataset_dir)

        ## read the data
        D =_read_csv(archive_path, feature_attrs=feature_attrs,
                     treatment_attrs=treatment_attrs,
                     target_attrs=target_attrs,
                     total_attrs=total_attrs,
                     csv_reader_args=csv_reader_args,
                     categ_as_strings=categ_as_strings,
                     header=header, all_num=all_num)

        joblib.dump(D, dataset_path, compress=9)
    elif not available and not download_if_missing:
        raise IOError("Data not found and `download_if_missing` is False")

    # get cached data
    try:
        D
    except NameError:
        D = joblib.load(dataset_path)

    # change columns into a table
    if as_frame:
        import pandas
        D.data = pandas.DataFrame(OrderedDict(zip(D.feature_names, D.data)))
    else:
        all_float = all((x.dtype.kind=="f") for x in D.data)
        if not all_float:
            for i, c in enumerate(D.data):
                D.data[i] = c.astype(object)
        D.data = np.column_stack(D.data)

    # final returned data
    ret = _prepare_final_data(D, shuffle, return_X_y)
    return ret
